
// GenerateExpr.java
// Andrew Davison, ad@coe.psu.ac.th, June 2023

/* This class is derived from the BNF grammar below but, instead
   of parsing an existing string, uses it as a guide to 
   randomly generate a tixy expression.

   Initially each branch (a "|") in the body of a rule
   has an equal liklihood of being chosen. However, when 
   the user scores an expression, the value is used to change
   the weights along the path taken through the parse tree,
   thereby affecting later choices.

   The management of these weights and the selection of
   branches through the parse tree is done by the Selector class.

   Usage:
     > compile GenerateExpr.java

   BNF rules for tixy expressions:

   expr  ::=
       "-" expr
       | expr  numOp  expr
       | "(" expr testOp expr ")"
       | "(" logicalExpr ")"
       | "(" expr bitOp expr ")"
       | literal
       | var
       | func

   numOp ::=  "+"  | "-"  | "*" | "/"  | "%"  | "**" | "^"
   
   testOp ::= ">" | "<" | ">=" | "<=" | "==" | "!="

   logicalExpr  ::=
      "!" expr
      | expr  logicalOp  expr
      | "true"
      | "false"

   logicalOp ::=  "&"  | "|" | "&&"  | "||"

   bitOp ::= "<<" | ">>" | ">>>"

   literal :: = double | int

   var ::= "t" | "i"  | "x"  | "y"

   func ::=
      "sin" "(" expr ")"
    | "cos" "(" expr ")"
    | "tan" "(" expr ")"
    | "atan2" "(" expr "," expr ")"
    | "random" "(" ")"
         :  // built-ins and functions in Expr4jExtras
         :
*/


import java.io.*;
import java.util.*;


public class GenerateExpr
{
  private static final int TOKS_LIMIT = 10;
    /* Number of tokens allowed in a string before the 
       the recursive generation of the expression is curtailed.
       This will mean that the final string may have more than
       TOKS_LIMIT tokens, but not too many more.
    */

  private static final int VBAD_SCORE = -2;
    /* the worst possible score, which is used to decide which
       expressions to store */


  /* IMPORTANT: if you change these ops and functions, then
     also update Selector.initSelector()  vvvvvvvvvvv
  */
  private static final String[] NUM_OPS =
             { "+", "-", "*",  "/", "%", "**", "^" };
                                         // two ways to write exponentation

  private static final String[] TEST_OPS =
                     { ">", "<", ">=", "<=", "==", "!=" };

  private static final String[] LOGICAL_OPS =
                     { "&", "|", "&&", "||" };

  private static final String[] BIT_OPS =
                     { "<<", ">>", ">>>" };

  private static final String[] VARS =
                     { "t", "t", "t", "i", "x", "y" };   
             // more t's makes choice of t more likely


  // FUNC_NMS and FUNC_ARITIES should have the same size
  private static final String[] FUNC_NMS =
       { "sin", "cos", "tan",
         "asin", "acos", "atan", "atan2",
         "sinh", "cosh", "tanh",
         "sqrt", "cbrt", "hypot",
         "abs", "min", "max", "random",
         "log", "log10", "log2", "exp",
         "floor", "ceil", "signum"
       };

  private static final int[] FUNC_ARITIES =
       { 1, 1, 1, 
         1, 1, 1, 2,
         1, 1, 1, 
         1, 1, 2,
         1, 2, 2, 0,
         1, 1, 1, 1,
         1, 1, 1
       };


  private Selector selector;
      // select a path through the parse rules

  private ArrayList<ArrayList<String>> exprsStore;
     // expressions generated during this run
  private ArrayList<Integer> scores;
     // scores assigned to these expressions


  private ArrayList<String> toks;
       // tokens used in the current expression
  private boolean exprTooLong;
  private int numVars;  // no. of variables in this expression

  private ArrayList<Pair<String,Integer>> choices;
      // rule:index pairs chosen during the generation

  private Random rand = new Random();


  public GenerateExpr()
  {
    selector = new Selector();
    choices = new ArrayList<>();
    exprsStore = new ArrayList<>();
    scores = new ArrayList<>();
  }  // end of GenerateExpr()


  public void close()
  /* When the application closes down, print the saved
     expressions with 'good' scores (> 0). Also print and
     save the current Selector weights associated with each rule.
  */
  {
    if (exprsStore.size() == 0)
      System.out.println("\nNo Expressions saved");
    else {
      System.out.println("\nGood Expressions:");
      for(int i=0; i < exprsStore.size(); i++) {
        if (scores.get(i) > 0)
          System.out.println((i+1) + ". (" + scores.get(i) +  "): " +
                     String.join("", exprsStore.get(i)));
      }
    }

    selector.print();
    selector.save();
  } // end of close()



  public String generate()
  /* Generate an expression as a series of tokens
     stored in toks. The branch choices made during
     the execution of the parsing rules are stored in
     the choices list.

     Store a 'good' expression in the exprsStore list, and 
     allocate it a default score of 0 until the actual score is
     set by the user.
  */
  {
    while (true) {
      // reset globals before expression generation
      exprTooLong = false;
      numVars = 0;
      toks = new ArrayList<String>();
      choices.clear();

      expr();  // execute parse rules

      if (worthSaving(toks, numVars)) {
        exprsStore.add(toks);
        scores.add(0);   // default until setScore() called
        return String.join("", toks);
      }
    }
  } // end of generate()


  private boolean worthSaving(ArrayList<String> toks, 
                                    int numVars)
  /* Various tests to determine if the expression
     (the toks list) is worth saving in exprsStore
  */
  { if (numVars == 0)      // no vars in the expression
      return false;
    else if (numVars > 5)  // too many vars
      return false;

    if (alreadyStored(toks, exprsStore))
      return false;

    return true;
  }  // end of worthSaving()


  private boolean alreadyStored(ArrayList<String> toks,
                     ArrayList<ArrayList<String>> exprsStore)
  // is toks already stored in exprsStore?
  {
    for(ArrayList<String> expr : exprsStore)
      if (toks.equals(expr)) 
        return true;
    return false;
  }  // end of alreadyStored()



  public void setScore(int score)
  /* Replaces the default score for an expression by the
     user's score unless that score is VBAD_SCORE in which
     case the expression is removed from exprsStore since
     it's deemed very bad.
  */
  {  scores.remove(scores.size()-1);   // remove default score
     if (score == VBAD_SCORE)
       exprsStore.remove(exprsStore.size()-1);   // remove expr from store
     else {  // set score and update choices
       scores.add(score);    // use real value
       selector.adjust(choices, score);
     }
  }  // end of setScore()



  // ------- parsing rules as functions ------------

  private void expr()
  /* expr  ::=
       "-" expr
       | expr numOp  expr
       | "(" expr testOp expr ")"
       | "(" logicalExpr ")"
       | "(" expr bitOp expr ")"
       | literal
       | var
       | func

     numOp ::=  "+"  | "-"  | "*" | "/"  | "%"  | "**" | "^"
   
     testOp ::= ">" | "<" | ">=" | "<=" | "==" | "!="
   
     bitOp ::= "<<" | ">>" | ">>>"
  */
  {
    if (toks.size() > TOKS_LIMIT)
      exprTooLong = true;

    int idx = exprTooLong ? randArrVal( new int[]{5,6}) :
                            selector.choose("expr");
       /* When the number of tokens exceeds the limit, 
          restrict the choices to rules that are not
          recursive. This will encourage the generation 
          to come to a close.
          Choices 5 and 6 are 'literal' and 'var'
       */
    choices.add( new Pair<String,Integer>("expr", idx));
             // save the choice as a (rule-name, choice-index) pair
    switch(idx) {
      case 0:
        toks.add("-"); expr();
        break;
      case 1:
        expr(); op("numOp", NUM_OPS); expr();
        break;
      case 2:
        toks.add("(");
        expr(); op("testOp", TEST_OPS); expr();
        toks.add(")");
        break;
      case 3:
        toks.add("(");
        logicalExpr();
        toks.add(")");
        break;
      case 4:
        toks.add("(");
        expr(); op("bitOp", BIT_OPS); expr();
        toks.add(")");
        break;
      case 5:
        literal(); break;
      case 6:
        var(); break;
      case 7:
        func(); break;
      default:
        System.out.println("Unknown expr index: " + idx);
    }
  }  // end of expr()



  private void op(String name, String[] ops)
  /* select an operation, save the choice, and add 
     the operation token to the tokens list */
  {
    int idx = selector.choose(name);
    if (idx != -1) {
      choices.add( new Pair<String,Integer>(name, idx));
           // save the choice as a (rule-name, choice-index) pair
      toks.add( ops[idx] );
    }
  }  // end of op()


  private int randArrVal(int[] arr)
  // randomly select a value from the array
  {
    int idx = rand.nextInt(arr.length);
    return arr[idx];
  }


  private void logicalExpr()
  /* logicalExpr  ::=
        "!" expr
        | expr  logicalOp  expr
        | "true"
        | "false"
   
     logicalOp ::=  "&"  | "|" | "&&"  | "||"
  */
  {
    int idx = selector.choose("logicalExpr");
    choices.add( new Pair<String,Integer>("logicalExpr", idx));
    switch(idx) {
      case 0:
        toks.add("!"); expr();
        break;
      case 1:
        expr(); op("logicalOp", LOGICAL_OPS); expr();
       break;
      case 2:
        toks.add("true"); break;
      case 3:
        toks.add("false"); break;
      default:
        System.out.println("Unknown logicalExpr index: " + idx);
    }
  }  // end of logicalExpr()



  private void literal()
  /* literal :: = double | int
     Since this code is generating tokens not parsing them,
     a literal is generated using random numbers.
  */
  { int idx = selector.choose("literal");
    choices.add( new Pair<String,Integer>("literal", idx));
    switch(idx) {
      case 0:
        randDouble(); break;
      case 1:
        randInt(); break;
      default:
        System.out.println("Unknown literal index: " + idx);
    }
  }  // end of literal()


  private void randDouble()
  // random double between 0 and 10, stored as a string with 1 dp
  {
    double val = 10 * rand.nextDouble();
    toks.add(String.format("%.1f", val));
  }


  private void randInt()
  // random integer 0 to 10, stored as a string
  { toks.add("" + rand.nextInt(10));  }


  private void var()
  //  var ::= "t" | "i"  | "x"  | "y"
  {
    int idx = selector.choose("var");
    choices.add( new Pair<String,Integer>("var", idx));
    toks.add( VARS[idx] );
    numVars++;
  }


  private void func()
  /* func ::= "sin" "(" expr ")"
            | "cos" "(" expr ")"
            | "tan" "(" expr ")"
            | "atan2" "(" expr "," expr ")"
            | "random" "(" ")"
                 :
     The functions listed in FUNC_NMS, and their 
     associated arities in FUNC_ARITIES are used to decide how
     many argument tokens are added after the function name token.
  */
  {
    int idx = selector.choose("func");
    if (idx == -1) {
      System.out.println("Unknown func index: " + idx);
      return;
    }

    choices.add( new Pair<String,Integer>("func", idx));

    toks.add(FUNC_NMS[idx]); toks.add("(");  // function name token

    // function argument tokens
    int arity = FUNC_ARITIES[idx];
    if (arity > 0) {
      while (arity > 1) {
        expr(); toks.add(",");
        arity--;
      }
      expr();
    }
    toks.add(")");
  }  // end of func()


}  // end of GenerateExpr class