
// Selector.java
// Andrew Davison, ad@coe.psu.ac.th, June 2023

/*  Each parsing rule for tixy expressions (see
    the comments for the GenerateExpr class)
    is represented by a map entry
    which maps a rule-name to a series of integers 
    representing the weights for the different "|" 
    branches in the body of the rule.

    For example the rule:
      numOp ::=  "+"  | "-"  | "*" | "/"  | "%"  | "**" | "^"
    becomes:
      numOp : [10, 10, 10, 10, 10, 10, 10]

    Initially all the weights are equal (START_WEIGHT (10)),
    but the user's score for a generated expression is used to 
    adjust the weights of all the choices used in the 
    generation of that expression.

    For example, if "t*t" gets a score of 2 then the "*" operator's
    weight will increase from 10 to 12. This increase will make it
    slightly more likely that "*" is chosen in the future.

    When the application opens, the current weights are
    loaded from SELECTOR_FNM (if it exists), and are saved
    back to that file when the application closes. 

    These weights represent a basic form of learning by
    TixyViewer since they influence the choice of future 
    expressions based on the scores for earlier expressions.

   Usage:
     > compile Selector.java
*/


import java.io.*;
import java.util.*;


public class Selector
{
  private static final int START_WEIGHT = 10;

  private static final String SELECTOR_FNM = 
      System.getProperty("user.home") + "/selectChoices.txt";
     // this directory is probably accessible for reading and writing


  private TreeMap<String, ArrayList<Integer>> selector;
      /* maps a rule-name to a series of integers representing
         the weights for the different "|" branches in the
         body of the rule
      */

  private Random rand = new Random();


  public Selector()
  // either load or initialize the selector weights
  {
    File f = new File(SELECTOR_FNM);
    if (f.exists() && !f.isDirectory())
      loadSelector(SELECTOR_FNM);
    else {
      System.out.println("No " +  SELECTOR_FNM + " found; initializing");
      initSelector();
    }
  }  // end of Selector()



  private void loadSelector(String fnm)
  /* The format of the input lines are:
        funcName:  int int int ...
     and blank lines and comment lines.
  */
  { 
    selector = new TreeMap<>();
    try {
      System.out.println("Loading selector file: " + fnm);

      BufferedReader br = new BufferedReader( new FileReader(fnm));
      String line;
      int numFuncs = 0;
      char ch;
      while((line = br.readLine()) != null) {
        line = line.trim();
        if (line.length() == 0)  // blank line
          continue;
        if (line.startsWith("//"))   // comment
          continue;

        String[] valStr = line.split("\\s+");
        if (valStr.length == 1)
          System.out.println("No weights for function: " + valStr[0]);
        else {
          ArrayList<Integer> weights = new ArrayList<>();
          for (int i=1; i < valStr.length; i++)
             weights.add( getInt(valStr[i]));
          selector.put(valStr[0], weights);
          numFuncs++;
        }
      }
      br.close();
      System.out.println("Loaded weights for " + numFuncs + " functions");
    } 
    catch (IOException e) 
    {  System.out.println("  " + e); }

  }  // end of loadSelector()


  private int getInt(String s)
  { 
    int num = 0;
    try {
      num = Integer.parseInt(s);
    }
    catch (NumberFormatException ex){ 
      System.out.println("Incorrect format for int " + s); 
    }
    return num;
  }  // end of getInt()



  private void initSelector()
  /* The number arguments come from the size of the OPS
     arrays in the ExprGen class, and the number of
     choices in the expr(), logicalExpr(), and
     literal() functions in that class
  */
  { selector = new TreeMap<>();
    buildRule("expr", 8);
    buildRule("numOp", 7);   // NUM_OPS length
    buildRule("testOp", 6);  // TEST_OPS
    buildRule("logicalExpr", 4);
    buildRule("logicalOp", 4);  // LOGICAL_OPS
    buildRule("bitOp", 3);  // BIT_OPS
    buildRule("literal", 2);
    buildRule("var", 6);  // VARS
    buildRule("func", 24);  // FUNC_NMS
      //              ^^^^^^^^^
      // these values MUST match the sizes of the arrays in ExprGen
  }  // end of initSelector()



  private void buildRule(String funcNm, int size)
  // create a list of weights that all use START_WEIGHT 
  {
    System.out.println("  Creating rule " +  funcNm + ": " + size);
    ArrayList<Integer> choices = new ArrayList<>();
    for( int i=0; i < size; i++)
      choices.add(START_WEIGHT);
    selector.put(funcNm, choices);
  }  // end of buildRule()



  public void adjust(ArrayList<Pair<String,Integer>> choices, int score)
  /*
     The choices made by expression generation are a series of
     pairs of the form (rule-name, choice-index). For each pair
     find the corresponding weight in the selector Map and modify
     it by adding the user's score to it. 

     The score can vary between -2 to 2 which will cause this choice's
     weight to decrease or increase, therby affecting how likely it is
     to be chosen in the future.
  */
  {
    for( Pair<String,Integer> choice : choices) {
      String ruleNm =  choice.getX();
      int idx = choice.getY();
      ArrayList<Integer> weights = selector.get(ruleNm);
      if (weights == null) 
        System.out.println("No rule " + ruleNm +  " in selector");
      else {
        int nWeight = weights.get(idx)+score;
           // modify the chosen weight with the score
        if (nWeight < 1)  
          nWeight = 1;   // don't let a weight go to 0
        weights.set(idx, nWeight);
        selector.put(ruleNm, weights);
      }
    }
  }  // end of adjust()



  public int choose(String ruleNm)
  /* Select a branch for the specified rule by summing
     all of its selector weights, and choosing a random number
     between 0 and the total. Match that to the index corresponding
     to that summed weight.

    For example the rule:
      numOp ::=  "+"  | "-"  | "*" | "/"  | "%"  | "**" | "^"
    becomes:
      numOp : [10, 10, 10, 10, 10, 10, 10]

    The total is 70. If the random number is 22, the index will
    == 2, which corresponds to the "*" operator.
  */
  {
    ArrayList<Integer> weights = selector.get(ruleNm);
    if (ruleNm == null) {
      System.out.println("No rule called " + ruleNm);
      return -1;
    }

    int total = weights.stream().mapToInt(Integer::intValue).sum();
    int boxVal = rand.nextInt(total);  // random number: 0 to total

    // which index corressponds to this summed weight?
    int sum = 0;
    for (int i = 0; i < weights.size(); i++) {
      sum += weights.get(i);
      if (boxVal <= sum)
        return i;
    }

    System.out.println("Error in selection for " + ruleNm);
    return -1;
  }  // end of choose()



  public void print()
  {
    System.out.println("\nSelector Rules:");
    for (Map.Entry<String, ArrayList<Integer>> entry :
                                        selector.entrySet())
      System.out.println("  " + entry.getKey() + ":" +
                              entry.getValue());
    System.out.println();
  }  // end of print()



  public void save()
  {  save(SELECTOR_FNM);  }


  public void save(String fnm)
  {
    System.out.println("Saving selector to file: " + fnm);
    try {
      int numFuncs = 0;
      PrintWriter pw = new PrintWriter(new FileWriter(fnm));
      for (String funcNm : selector.keySet()) {
        pw.print(funcNm);
        ArrayList<Integer> weights = selector.get(funcNm);
        for (Integer w : weights)
           pw.print(" " + w);
        pw.println();
        numFuncs++;
      }
      // System.out.println("Written " + numFuncs + " rules");
      pw.close();
    }
    catch (IOException e) 
    {  System.out.println("  " + e); }

  }  // end of save()


}  // end of Selector class
