
// TixyViewer.java
// Andrew Davison, Jan 2022, ad@fivedots.coe.psu.ac.th

/* TixyViewer draws a grid of cells inside a JPanel
   to display the current tixy expression. The expression
   is NOT written by the user but generated by the 
   GenerateExpr class.

   main() places TixyViewer inside a JFrame which includes
   a textfield that displays the expression string, a button for
   sending that expressionn to the tixyland website 
   (https://tixy.land) for 'official' rendering, and
   five scoring buttons.

   Pressing a scoring button, stores the expression so it
   can be listed when the application closes if it has
   a score > 0. Also, a new tixy expression is automatically
   generated and displayed.

   Usage:
     > compile TixyViewer.java
     > run TixyViewer
*/

import java.io.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.awt.geom.*;
import java.util.*;

import java.net.*;
import java.nio.charset.*;

import net.objecthunter.exp4j.*;
import net.objecthunter.exp4j.extras.*;



public class TixyViewer extends JPanel implements Runnable
{
  private static final int VBAD_SCORE = -2;  // worst score

  private static final int PSIZE = 384;     // pixel size of this panel
  private static final int DELAY = 100;     // ms pause between each grid redraw
  private static final int NUM_BOXES = 16;   // along x- and y- axes
  private static final int BOX_SIZE = PSIZE/NUM_BOXES;   // size in pixels

  private double[][] grid = new double[NUM_BOXES][NUM_BOXES];
         // used to visualize the tixy expression

  private GenerateExpr genExpr;
  private String exprStr; 

  private boolean isValidMath, isRunning, isInteresting;
     /* isValidMath: is the expression a valid math expression?
        isRunning: is the grid being regularly redrawn?
        isInteresting: is the rendering of the expression an
                       interesting-looking image?
     */
                  
  private JTextField exprTF;  // shows the expression string


  public TixyViewer(JTextField tf)
  {
    // generate an initial expression
    exprTF = tf;
    genExpr = new GenerateExpr();
    exprStr = genExpr.generate();
    exprTF.setText(" " + exprStr);
    isValidMath = true;   // hope for the best
    isInteresting = false;

    setBackground(Color.BLACK);
    setPreferredSize(new Dimension(PSIZE, PSIZE));
    isRunning = true;
    new Thread(this).start();
  }   // end of TixyViewer()



  public void score(int val)
  /* record the user's score for the current expression,
     then generate and display a new expression
  */
  {
     genExpr.setScore(val);

     exprStr = genExpr.generate();
     exprTF.setText(" " + exprStr);
     exprTF.repaint();
     isValidMath = true;   // hope for the best
     isInteresting = false;
  } // end of score()



  public void run()
  // update/redraw/pause loop
  {
    double time = 0;
    while(isRunning) {
      updateGrid(time);
      if (!isInteresting) {
        while (!isInterestingGrid()) {
          /* keep generating a new expression until its grid
             is deemed interesting. Uninteresting expressions
             are still stored, but given a very bad score, 
             and not drawn in the panel */
          score(VBAD_SCORE);
          updateGrid(time);
        }
        isInteresting = true;  // since grid is 'good'
      }
      repaint();
      pause(DELAY);
      time = time + ((double)DELAY)/1000;   // time is stored in seconds
    }
    genExpr.close();

    System.exit(0);  // close appl.
  } // end of run()


  public void close()
  {  isRunning = false;  }


  private void updateGrid(double t)
  // apply the tixy expression to every cell of the grid
  {
    for (int x=0; x < NUM_BOXES; x++) {
      for (int y=0; y < NUM_BOXES; y++) {
        int i = (y * 16) + x;
        grid[y][x] =  tixyFn(t, i, x, y);
      }
    }
  }  // end of updateGrid()


  private double tixyFn(double t, double i, double x, double y)
  /* Use all the Expr4j functions and operators and
     extra ones defined in Expr4jExtras.
     Expr4j (https://gitlab.com/riddler_arg/exp4j) is a library to 
     parse and evaluate mathematical expression strings.
  */
  {
    double result = 1;   // drawn as a full-size white circle
    if (isValidMath) {
      try {
        Expression expr = new ExpressionBuilder(exprStr)
               .variables("t", "i", "x", "y")
               .operators(OperatorsComparison.getOperators())
               .operators(Expr4jExtras.ops)
               .functions(FunctionsBoolean.getFunctions())
               .functions(FunctionsMisc.getFunctions())
               .functions(Expr4jExtras.funcs)
               .build();

        // set the variables used in this expression
        if (expr.containsVariable("t"))
          expr.setVariable("t", t);
        if (expr.containsVariable("i"))
          expr.setVariable("i", i);
        if (expr.containsVariable("x"))
          expr.setVariable("x", x);
        if (expr.containsVariable("y"))
          expr.setVariable("y", y);

        result = expr.evaluate();
      }
      catch(IllegalArgumentException e) 
      {  System.out.println(exprStr + " : " + e.getMessage());
         isValidMath = false;
      }
      catch(ArithmeticException e) 
      {
         // System.out.println(e.getMessage());
              // usually division by zero; use a default grid value
         result = 1;   // a full-size white circle
      }
    }
    return result;
  }  // end of tixyFn()


  private boolean isInterestingGrid()
  /* apply a series of tests to the grid values to
     decide if the rendered grid image will be interesting
  */
  {
    double sum = 0;
    for (int x=0; x < NUM_BOXES; x++)
      for (int y=0; y < NUM_BOXES; y++) {
        double sz = Math.min(Math.max(Math.abs(grid[y][x]), 0), 1);  // 0 to 1
        sum += sz;
      }

    if (Double.isNaN(sum))
      return false;
/*
    if (exprStr.matches(".*\\bt\\b.*"))  // uses t variable?
      return true;
*/
    if (sum < 3.0)  
      // very close to 0; i.e. the grid is totally black
      return false;
    else if (sum > ((NUM_BOXES*NUM_BOXES)-3))   
      // very close to 1/-1; i.e. the grid is totally white/red circles
      return false;

    return true;
  }  // end of isInterestingGrid()


  private void pause(int ms)
  {
    try {
      Thread.sleep(ms);
    }
    catch(InterruptedException ex) { }
  }  // end of pause()




  public void paintComponent(Graphics g)
  {
    super.paintComponent(g); 
    Graphics2D g2d = (Graphics2D) g;

    g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                         RenderingHints.VALUE_ANTIALIAS_ON);
    g2d.setRenderingHint(RenderingHints.KEY_RENDERING,
                         RenderingHints.VALUE_RENDER_QUALITY);

    for (int x=0; x < NUM_BOXES; x++) {
      for (int y=0; y < NUM_BOXES; y++) {
        double boxVal = grid[y][x];
        g2d.setColor( (boxVal < 0) ? Color.RED : Color.WHITE);

        double sz = Math.min(Math.max(Math.abs(boxVal), 0), 1);
        double width = sz*BOX_SIZE;
        double offset = (BOX_SIZE - width)/2.0;
        double xc = (x*BOX_SIZE) + offset;   // top-left (x,y) for the circle
        double yc = (y*BOX_SIZE) + offset;

        Ellipse2D circle = new Ellipse2D.Double(xc, yc, width, width);
        g2d.fill(circle);
      }
    }
  } // end of paintComponent()



  public void openInBrowser()
  /* Send the expression string to the tixyland site for 'official'
     rendering. This may not look the same since the site doesn't 
     support all of the operations/functions in Expr4J
  */
  {
    String urlStr = "https://tixy.land/?code=" + encodeExpr(exprStr);

    Desktop dt = Desktop.isDesktopSupported() ? Desktop.getDesktop() : null;
    if ((dt != null) && dt.isSupported(Desktop.Action.BROWSE)) {
      try {
        dt.browse(new URI(urlStr));
      } 
      catch (IOException | URISyntaxException e) 
      {  System.out.println(e); }
    }
    else {  // Try MS Windows 'start'
      Runtime rt = Runtime.getRuntime();
      try {
        String[] cmds = {"cmd.exe","/c","start", urlStr};
        rt.exec(cmds);
      } 
      catch (IOException e) 
      {  System.out.println(e); }
    }
  }  // end of openInBrowser()


  private String encodeExpr(String s) 
  {
    try {
      return URLEncoder.encode(s, StandardCharsets.UTF_8.toString());
    }
    catch (UnsupportedEncodingException e) {
      System.out.println(e);
      return s;
    }
  }


  // -----------------------------------------


  public static void main(String[] args)
  {
    JFrame f = new JFrame("Tixy Viewer");
    Container c = f.getContentPane();

    // ------ expression textfield -----
    JPanel p1 = new JPanel();
    JTextField exprTF = new JTextField(25);
    exprTF.setFont( new Font("Verdana", Font.PLAIN, 14));
    exprTF.setEditable(false);

    // the grid visualization in a JPanel
    TixyViewer tixy = new TixyViewer(exprTF);
    c.add(tixy, BorderLayout.CENTER);

    // ----- button controls ------

    JPanel p2 = new JPanel();
    JButton openB = new JButton("Open in Browser");
    openB.addActionListener( (ev) -> { tixy.openInBrowser(); });
    p2.add(openB);

    JPanel ctrls = new JPanel();
    ctrls.setLayout( new BorderLayout());
    ctrls.add(exprTF, BorderLayout.NORTH);
    ctrls.add(p2, BorderLayout.CENTER);
    ctrls.add(buildScoring(tixy), BorderLayout.SOUTH);

    c.add( ctrls, BorderLayout.SOUTH);

    // f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    f.addWindowListener(new WindowAdapter() {
       public void windowClosing(WindowEvent e) 
       {  tixy.close(); }
    });
    f.pack();
    f.setResizable(false);
    f.setVisible(true);
  }  // end of main()



  private static JPanel buildScoring(TixyViewer tixy)
  /* five scoring buttons in a row: 
     the scores go from VBAD_SCORE (-2) to 2
  */
  {
    JPanel pButs = new JPanel();
    pButs.setLayout( new FlowLayout());
   
    JButton b0 = new JButton("V.Bad");
    b0.addActionListener( (ev) -> { tixy.score(VBAD_SCORE); });
    pButs.add(b0);
   
    JButton b1 = new JButton("Bad");
    b1.addActionListener( (ev) -> { tixy.score(-1); });
    pButs.add(b1);
   
    JButton b2 = new JButton("Ok");
    b2.addActionListener( (ev) -> { tixy.score(0); });
    pButs.add(b2);
   
    JButton b3 = new JButton("Good");
    b3.addActionListener( (ev) -> { tixy.score(1); });
    pButs.add(b3);
   
    JButton b4 = new JButton("V.Good");
    b4.addActionListener( (ev) -> { tixy.score(2); });
    pButs.add(b4);
   
    return pButs;
  }  // end of buildScoring()


}  // end of TixyViewer class
